import re
import warnings
from collections import Counter
from collections.abc import ValuesView
from typing import Optional, Union

import numpy as np
import pandas as pd
from great_tables import GT
from tabulate import tabulate

from pyfixest.estimation.feiv_ import Feiv
from pyfixest.estimation.feols_ import Feols
from pyfixest.estimation.fepois_ import Fepois
from pyfixest.estimation.FixestMulti_ import FixestMulti
from pyfixest.report.utils import _relabel_expvar
from pyfixest.utils.dev_utils import _select_order_coefs

ModelInputType = Union[
    FixestMulti, Feols, Fepois, Feiv, list[Union[Feols, Fepois, Feiv]]
]


def etable(
    models: ModelInputType,
    type: str = "gt",
    signif_code: Optional[list] = None,
    coef_fmt: str = "b \n (se)",
    custom_stats: Optional[dict] = None,
    custom_model_stats: Optional[dict] = None,
    keep: Optional[Union[list, str]] = None,
    drop: Optional[Union[list, str]] = None,
    exact_match: Optional[bool] = False,
    labels: Optional[dict] = None,
    cat_template: Optional[str] = None,
    show_fe: Optional[bool] = True,
    show_se_type: Optional[bool] = True,
    felabels: Optional[dict] = None,
    notes: str = "",
    model_heads: Optional[list] = None,
    head_order: Optional[str] = "dh",
    file_name: Optional[str] = None,
    **kwargs,
) -> Union[pd.DataFrame, str, None]:
    r"""
    Generate a table summarizing the results of multiple regression models.
    It supports various output formats including html (via great tables),  markdown, and LaTeX.

    Parameters
    ----------
    models : A supported model object (Feols, Fepois, Feiv, FixestMulti) or a list of
            Feols, Fepois & Feiv models.
        The models to be summarized in the table.
    type : str, optional
        Type of output. Either "df" for pandas DataFrame, "md" for markdown,
        "gt" for great_tables, or "tex" for LaTeX table. Default is "gt".
    signif_code : list, optional
        Significance levels for the stars. Default is None, which sets [0.001, 0.01, 0.05].
        If None, no stars are printed.
    coef_fmt : str, optional
        The format of the coefficient (b), standard error (se), t-stats (t), and
        p-value (p). Default is `"b \n (se)"`.
        Spaces ` `, parentheses `()`, brackets `[]`, newlines `\n` are supported.
    custom_stats: dict, optional
        A dictionary of custom statistics that can be used in the coef_fmt string to be displayed
        in the coefficuent cells analogously to "b", "se" etc. The keys are the names of the custom
        statistics, and the values are lists of lists, where each inner list contains the custom
        statistic values for all coefficients each model.
        Note that "b", "se", "t", or "p" are reserved and cannot be used as keys.
    custom_model_stats: dict, optional
        A dictionary of custom model statistics or model information displayed in a new line in the
        bottom panel of the table. The keys are the names of the statistics (i.e. entry in the first column)
        and the values are a lists of the same length as the number of models. Default is None.
    keep: str or list of str, optional
        The pattern for retaining coefficient names. You can pass a string (one
        pattern) or a list (multiple patterns). Default is keeping all coefficients.
        You should use regular expressions to select coefficients.
            "age",            # would keep all coefficients containing age
            r"^tr",           # would keep all coefficients starting with tr
            r"\\d$",          # would keep all coefficients ending with number
        Output will be in the order of the patterns.
    drop: str or list of str, optional
        The pattern for excluding coefficient names. You can pass a string (one
        pattern) or a list (multiple patterns). Syntax is the same as for `keep`.
        Default is keeping all coefficients. Parameter `keep` and `drop` can be
        used simultaneously.
    exact_match: bool, optional
        Whether to use exact match for `keep` and `drop`. Default is False.
        If True, the pattern will be matched exactly to the coefficient name
        instead of using regular expressions.
    labels: dict, optional
        A dictionary to relabel the variables. The keys in this dictionary are the
        original variable names, which correspond to the names stored in the `_coefnames`
        attribute of the model. The values in the dictionary are the new names you want
        to assign to these variables.
        Note that interaction terms will also be relabeled using the labels of the individual variables.
        The command is applied after the `keep` and `drop` commands.
    cat_template: str, optional
        Template to relabel categorical variables. None by default, which applies no relabeling.
        Other options include combinations of "{variable}" and "{value}", e.g. "{variable}::{value}"
        to mimic fixest encoding. But "{variable}--{value}" or "{variable}{value}" or just "{value}"
        are also possible.
    show_fe: bool, optional
        Whether to show the rows with fixed effects markers. Default is True.
    show_se_type: bool, optional
        Whether to show the rows with standard error type. Default is True.
    felabels: dict, optional
        A dictionary to relabel the fixed effects. Only needed if you want to relabel
        the FE lines with a different label than the one specied for the respective
        variable in the labels dictionary.
        The command is applied after the `keep` and `drop` commands.
    digits: int
        The number of digits to round to.
    thousands_sep: bool, optional
        The thousands separator. Default is False.
    scientific_notation: bool, optional
        Whether to use scientific notation. Default is True.
    scientific_notation_threshold: int, optional
        The threshold for using scientific notation. Default is 10_000.
    notes: str, optional
        Custom table notes. Default shows the significance levels and the format of
        the coefficient cell.
    model_heads: list, optional
        Add custom headlines to models when output as df or latex. Length of list
        must correspond to number of models. Default is None.
    head_order: str, optional
        String to determine the display of the table header when output as df or latex.
        Allowed values are "dh", "hd", "d", "h", or "". When head_order is "dh",
        the dependent variable is displayed first, followed by the custom model_heads
        (provided the user has specified them). With "hd" it is the other way around.
        When head_order is "d", only the dependent variable and model numbers are displayed
        and with "" only the model numbers. Default is "dh".
    file_name: str, optional
        The name/path of the file to save the LaTeX table to. Default is None.

    Returns
    -------
    pandas.DataFrame
        A styled DataFrame with the coefficients and standard errors of the models.
        When output is "tex", the LaTeX code is returned as a string.

    Examples
    --------
    For more examples, take a look at the [regression tables and summary statistics vignette](https://py-econometrics.github.io/pyfixest/table-layout.html).

    ```{python}
    import pyfixest as pf

    # load data
    df = pf.get_data()
    fit1 = pf.feols("Y~X1 + X2 | f1", df)
    fit2 = pf.feols("Y~X1 + X2 | f1 + f2", df)

    pf.etable([fit1, fit2])
    ```
    """
    if signif_code is None:
        signif_code = [0.001, 0.01, 0.05]
    assert isinstance(signif_code, list) and len(signif_code) == 3, (
        "signif_code must be a list of length 3"
    )
    if signif_code:
        assert all([0 < i < 1 for i in signif_code]), (
            "All values of signif_code must be between 0 and 1"
        )
    if signif_code:
        assert signif_code[0] < signif_code[1] < signif_code[2], (
            "signif_code must be in increasing order"
        )

    cat_template = "" if cat_template is None else cat_template

    models = _post_processing_input_checks(models)

    if labels is None:
        labels = {}
    if custom_stats is None:
        custom_stats = dict()
    if keep is None:
        keep = []
    if drop is None:
        drop = []

    if custom_stats:
        assert isinstance(custom_stats, dict), "custom_stats must be a dict"
        for key in custom_stats:
            assert isinstance(custom_stats[key], list), (
                "custom_stats values must be a list"
            )
            assert len(custom_stats[key]) == len(models), (
                f"custom_stats {key} must have the same number as models"
            )

    assert type in [
        "df",
        "tex",
        "md",
        "html",
        "gt",
    ], "type must be either 'df', 'md', 'html', 'gt' or 'tex'"

    if model_heads is not None:
        assert len(model_heads) == len(models), (
            "model_heads must have the same length as models"
        )

    # Check if head_order is allowed string & remove h when no model_heads provided
    assert head_order in [
        "dh",
        "hd",
        "d",
        "h",
        "",
    ], "head_order must be one of 'd', 'h', 'dh', 'hd', ''"
    if model_heads is None and "h" in head_order:
        head_order = head_order.replace("h", "")

    # Check if custom_model_stats is a dictionary and the provided lists have the same length as models
    if custom_model_stats is not None:
        assert isinstance(custom_model_stats, dict), "custom_model_stats must be a dict"
        for stat, values in custom_model_stats.items():
            assert isinstance(stat, str), "custom_model_stats keys must be strings"
            assert isinstance(values, list), "custom_model_stats values must lists"
            assert len(values) == len(models), (
                "lists in custom_model_stats values must have the same length as models"
            )

    dep_var_list = []
    nobs_list = []
    fixef_list: list[str] = []
    n_coefs = []
    se_type_list = []
    r2_list = []
    adj_r2_list = []
    r2_within_list = []

    # Define code for R2, interaction & line break depending on output type
    if type in ["gt", "html"]:
        interactionSymbol = " &#215; "
        R2code = "R<sup>2</sup>"
        adj_R2_code = "Adj. R<sup>2</sup>"
        R2_within_code = "R<sup>2</sup> Within"
        lbcode = "<br>"
    elif type == "tex":
        interactionSymbol = " $\\times$ "
        R2code = "$R^2$"
        adj_R2_code = "Adj. $R^2$"
        R2_within_code = "$R^2$ Within"
        lbcode = r"\\"
    else:
        interactionSymbol = " x "
        R2code = "R2"
        adj_R2_code = "Adj. R2"
        R2_within_code = "R2 Within"
        lbcode = "\n"

    for model in models:
        dep_var_list.append(model._depvar)
        n_coefs.append(len(model._coefnames))

        _nobs_kwargs = kwargs.copy()
        _nobs_kwargs["integer"] = True
        _nobs_kwargs["scientific_notation"] = False
        nobs_list.append(_number_formatter(model._N, **_nobs_kwargs))

        if not np.isnan(model._r2):
            r2_list.append(_number_formatter(model._r2, **kwargs))
        else:
            r2_list.append("-")

        if not np.isnan(model._adj_r2):
            adj_r2_list.append(_number_formatter(model._adj_r2, **kwargs))
        else:
            adj_r2_list.append("-")

        if not np.isnan(model._r2_within):
            r2_within_list.append(_number_formatter(model._r2_within, **kwargs))
        else:
            r2_within_list.append("-")

        if model._vcov_type == "CRV":
            se_type_list.append("by: " + "+".join(model._clustervar))
        else:
            se_type_list.append(model._vcov_type)

        if model._fixef is not None and model._fixef != "0":
            fixef_list += model._fixef.split("+")

    # find all fixef variables when the user does not want to hide the FE rows
    if show_fe:
        # drop "" from fixef_list
        fixef_list = [x for x in fixef_list if x]
        # keep only unique values
        fixef_list = list(set(fixef_list))
        n_fixef = len(fixef_list)
    else:
        fixef_list = []
        n_fixef = 0

    # First create a dataframe for the model stats such as R2, nobs, etc.
    model_stats_df = pd.DataFrame()
    if custom_model_stats is not None:
        for stat, values in custom_model_stats.items():
            model_stats_df[stat] = values
    model_stats_df["Observations"] = nobs_list
    if show_se_type:
        model_stats_df["S.E. type"] = se_type_list
    model_stats_df[R2code] = r2_list
    n_model_stats = model_stats_df.shape[1]
    if any(x != "-" for x in r2_within_list):
        model_stats_df[R2_within_code] = r2_within_list
    else:
        model_stats_df[adj_R2_code] = adj_r2_list
    # Transpose
    model_stats_df = model_stats_df.T

    # Create a dataframe for the Fixed Effects markers
    fe_df = pd.DataFrame()
    # when at least one model has a fixed effect & the user wants to show them
    if fixef_list:
        for fixef in fixef_list:
            # check if not empty string
            if fixef:
                for i, model in enumerate(models):
                    if (
                        model._fixef is not None
                        and fixef in model._fixef.split("+")
                        and not model._use_mundlak
                    ):
                        fe_df.loc[i, fixef] = "x"
                    else:
                        fe_df.loc[i, fixef] = "-"
        # Sort by model
        fe_df.sort_index(inplace=True)
        # Transpose
        fe_df = fe_df.T
    else:
        show_fe = False

    # Finally, collect & format estimated coefficients and standard errors etc.
    coef_fmt_elements, coef_fmt_title = _parse_coef_fmt(coef_fmt, custom_stats)
    etable_list = []
    for i, model in enumerate(models):
        model_tidy_df = model.tidy()
        model_tidy_df.reset_index(
            inplace=True
        )  # If rounding here and p = 0.0499, it will be rounded to 0.05 and miss threshold.
        model_tidy_df["stars"] = (
            np.where(
                model_tidy_df["Pr(>|t|)"] < signif_code[0],
                "***",
                np.where(
                    model_tidy_df["Pr(>|t|)"] < signif_code[1],
                    "**",
                    np.where(model_tidy_df["Pr(>|t|)"] < signif_code[2], "*", ""),
                ),
            )
            if signif_code
            else ""
        )
        model_tidy_df[coef_fmt_title] = ""
        for element in coef_fmt_elements:
            if element == "b":
                model_tidy_df[coef_fmt_title] += (
                    model_tidy_df["Estimate"].apply(_number_formatter, **kwargs)
                    + model_tidy_df["stars"]
                )
            elif element == "se":
                model_tidy_df[coef_fmt_title] += model_tidy_df["Std. Error"].apply(
                    _number_formatter, **kwargs
                )
            elif element == "t":
                model_tidy_df[coef_fmt_title] += model_tidy_df["t value"].apply(
                    _number_formatter, **kwargs
                )
            elif element == "p":
                model_tidy_df[coef_fmt_title] += model_tidy_df["Pr(>|t|)"].apply(
                    _number_formatter, **kwargs
                )
            elif element in custom_stats:
                assert len(custom_stats[element][i]) == len(
                    model_tidy_df["Estimate"]
                ), (
                    f"custom_stats {element} has unequal length to the number of coefficients in model_tidy_df {i}"
                )
                model_tidy_df[coef_fmt_title] += pd.Series(
                    custom_stats[element][i]
                ).apply(_number_formatter, **kwargs)
            elif element == "\n":  # Replace output specific code for newline
                model_tidy_df[coef_fmt_title] += lbcode
            else:
                model_tidy_df[coef_fmt_title] += element
        model_tidy_df[coef_fmt_title] = pd.Categorical(model_tidy_df[coef_fmt_title])
        model_tidy_df = model_tidy_df[["Coefficient", coef_fmt_title]]
        model_tidy_df = pd.melt(
            model_tidy_df,
            id_vars=["Coefficient"],
            var_name="Metric",
            value_name=f"est{i + 1}",
        )
        model_tidy_df = model_tidy_df.drop("Metric", axis=1).set_index("Coefficient")
        etable_list.append(model_tidy_df)

    res = pd.concat(etable_list, axis=1)
    if keep or drop:
        idxs = _select_order_coefs(res.index.tolist(), keep, drop, exact_match)
    else:
        idxs = res.index
    res = res.loc[idxs, :].reset_index()
    # a lot of work to replace the NaNs with empty strings
    # reason: "" not a level of the category, might lead to a pandas error
    for column in res.columns:
        if (
            isinstance(res[column].dtype, pd.CategoricalDtype)
            and "" not in res[column].cat.categories
        ):
            res[column] = res[column].cat.add_categories([""])

        # Replace NA values with the empty string
        res[column] = res[column].fillna("")

    res.rename(columns={"Coefficient": "index"}, inplace=True)
    res.set_index("index", inplace=True)

    # Move the intercept row (if there is one) to the bottom of the table
    if "Intercept" in res.index:
        intercept_row = res.loc["Intercept"]
        res = res.drop("Intercept")
        res = pd.concat([res, pd.DataFrame([intercept_row])])

    # Relabel variables
    if (labels != {}) or (cat_template != ""):
        # Relabel dependent variables
        dep_var_list = [labels.get(k, k) for k in dep_var_list]

        # Relabel explanatory variables
        res_index = res.index.to_series()
        res_index = res_index.apply(
            lambda x: _relabel_expvar(x, labels or {}, interactionSymbol, cat_template)
        )
        res.set_index(res_index, inplace=True)

    # Relabel fixed effects
    if show_fe:
        if felabels is None:
            felabels = dict()
        if labels is None:
            labels = dict()
        # When the user provides a dictionary for fixed effects, then use it
        # When a corresponsing variable is not in the felabel dictionary, then use the labels dictionary
        # When in neither then just use the original variable name
        fe_index = fe_df.index.to_series()
        fe_index = fe_index.apply(lambda x: felabels.get(x, labels.get(x, x)))
        fe_df.set_index(fe_index, inplace=True)

    model_stats_df.columns = res.columns
    if show_fe:
        fe_df.columns = res.columns

    depvars = pd.DataFrame({"depvar": dep_var_list}).T
    depvars.columns = res.columns

    if type == "df":
        res_all = pd.concat([depvars, res, fe_df, model_stats_df])
        return res_all
    elif type == "md":
        res_all = pd.concat([depvars, res, fe_df, model_stats_df]).reset_index()
        # Generate notes string if user has not provided any
        if notes is None:
            if signif_code:
                notes = f"Significance levels: * p < {signif_code[2]}, ** p < {signif_code[1]}, *** p < {signif_code[0]}"
            else:
                notes = f"Format of coefficient cell: {coef_fmt_title}"
        res_all = _tabulate_etable_md(
            df=res_all,
            n_coef=res.shape[0],
            n_fixef=n_fixef,
            n_models=len(models),
            n_model_stats=n_model_stats,
        )
        print(res_all)
        print(notes)
        return None

    elif type in ["tex", "gt"]:
        # Prepare Multiindex for columns
        id_dep = dep_var_list  # depvars
        id_head = [""] * len(models) if model_heads is None else model_heads
        id_num = [f"({s})" for s in range(1, len(models) + 1)]  # model numbers

        # Concatenate the dataframes for coefficients, fixed effects, and model stats
        # and add keys identifying the three parts which will allow make_table
        # to format the table correctly inserting line between the parts
        res_all = pd.concat([res, fe_df, model_stats_df], keys=["coef", "fe", "stats"])

        # When no depvars & headlines should be displayed then use simple index
        # otherwise generate MultiIndex & determine order of index levels as specified by head_order
        if head_order == "":
            res_all.columns = pd.Index(id_num)
        else:
            cindex = [{"h": id_head, "d": id_dep}[c] for c in head_order] + [id_num]
            res_all.columns = pd.MultiIndex.from_arrays(cindex)

        # Generate generic note string if none is provided
        if notes == "":
            if type == "gt":
                notes = (
                    f"Significance levels: * p < {signif_code[2]}, ** p < {signif_code[1]}, *** p < {signif_code[0]}. "
                    + f"Format of coefficient cell:\n{coef_fmt_title}"
                )
            elif type == "tex":
                notes = (
                    f"Significance levels: $*$ p $<$ {signif_code[2]}, $**$ p $<$ {signif_code[1]}, $***$ p $<$ {signif_code[0]}. "
                    + f"Format of coefficient cell: {coef_fmt_title}"
                )
        return make_table(
            res_all, type=type, notes=notes, rgroup_display=False, **kwargs
        )

    return None


def summary(models: ModelInputType, digits: int = 3) -> None:
    """
    Print a summary of estimation results for each estimated model.

    For each model, this method prints a header indicating the fixed-effects and the
    dependent variable, followed by a table of coefficient estimates with standard
    errors, t-values, and p-values.

    Parameters
    ----------
    models : A supported model object (Feols, Fepois, Feiv, FixestMulti) or a list of
            Feols, Fepois & Feiv models.
    digits : int, optional
        The number of decimal places to round the summary statistics to. Default is 3.

    Returns
    -------
    None

    Examples
    --------
    ```{python}
    import pyfixest as pf

    # load data
    df = pf.get_data()
    fit1 = pf.feols("Y~X1 + X2 | f1", df)
    fit2 = pf.feols("Y~X1 + X2 | f1 + f2", df)
    fit3 = pf.feols("Y~X1 + X2 | f1 + f2 + f3", df)

    pf.summary([fit1, fit2, fit3])
    ```
    """
    models = _post_processing_input_checks(models)

    for fxst in list(models):
        depvar = fxst._depvar

        df = fxst.tidy().round(digits)

        if fxst._method == "feols":
            estimation_method = "IV" if fxst._is_iv else "OLS"
        elif fxst._method == "fepois":
            estimation_method = "Poisson"
        elif fxst._method == "twfe":
            estimation_method = "TWFE"
        elif fxst._method == "did2s":
            estimation_method = "DID2S"
        elif "quantreg" in fxst._method:
            estimation_method = f"quantreg: q = {fxst._quantile}"  # type: ignore
        else:
            raise ValueError("Unknown estimation method.")
        print("###")
        print("")
        print("Estimation: ", estimation_method)
        depvar_fixef = f"Dep. var.: {depvar}"
        if fxst._fixef is not None:
            if not fxst._use_mundlak:
                depvar_fixef += f", Fixed effects: {fxst._fixef}"
            else:
                depvar_fixef += f", Mundlak: by {fxst._fixef}"
        print(depvar_fixef)
        if fxst._sample_split_value != "all":
            split = f"sample: {fxst._sample_split_var} = {fxst._sample_split_value}"
            print(split)
        print("Inference: ", fxst._vcov_type_detail)
        print("Observations: ", fxst._N)
        print("")
        print(df.to_markdown(floatfmt=f".{digits}f"))
        print("---")

        to_print = ""

        if not np.isnan(fxst._rmse):
            to_print += f"RMSE: {np.round(fxst._rmse, digits)} "
        if not np.isnan(fxst._r2):
            to_print += f"R2: {np.round(fxst._r2, digits)} "
        if not np.isnan(fxst._r2_within):
            to_print += f"R2 Within: {np.round(fxst._r2_within, digits)} "
        if fxst.deviance is not None:
            to_print += f"Deviance: {np.round(fxst.deviance[0], digits)} "

        print(to_print)


def _post_processing_input_checks(
    models: ModelInputType,
    check_duplicate_model_names: bool = False,
    rename_models: Optional[dict[str, str]] = None,
) -> list[Union[Feols, Fepois, Feiv]]:
    """
    Perform input checks for post-processing models.

    Parameters
    ----------
        models : Union[List[Union[Feols, Fepois, Feiv]], FixestMulti]
                The models to be checked. This can either be a list of models
                (Feols, Fepois, Feiv) or a single FixestMulti object.
        check_duplicate_model_names : bool, optional
                Whether to check for duplicate model names. Default is False.
                Mostly used to avoid overlapping models in plots created via
                pf.coefplot() and pf.iplot().
        rename_models : dict, optional
                A dictionary to rename the models. The keys are the original model names
                and the values are the new model names.

    Returns
    -------
        List[Union[Feols, Fepois]]
            A list of checked and validated models. The returned list contains only
            Feols and Fepois types.

    Raises
    ------
        TypeError: If the models argument is not of the expected type.

    """
    models_list: list[Union[Feols, Fepois, Feiv]] = []

    if isinstance(models, (Feols, Fepois, Feiv)):
        models_list = [models]
    elif isinstance(models, FixestMulti):
        models_list = models.to_list()
    elif isinstance(models, (list, ValuesView)):
        if all(isinstance(m, (Feols, Fepois, Feiv)) for m in models):
            models_list = models
        else:
            raise TypeError(
                "All elements in the models list must be instances of Feols, Feiv, or Fepois."
            )
    else:
        raise TypeError("Invalid type for models argument.")

    if check_duplicate_model_names or rename_models is not None:
        all_model_names = [model._model_name for model in models_list]

    if check_duplicate_model_names:
        # create model_name_plot attribute to differentiate between models with the
        # same model_name / model formula
        for model in models_list:
            model._model_name_plot = model._model_name

        counter = Counter(all_model_names)
        duplicate_model_names = [item for item, count in counter.items() if count > 1]

        for duplicate_model in duplicate_model_names:
            duplicates = [
                model for model in models_list if model._model_name == duplicate_model
            ]
            for i, model in enumerate(duplicates):
                model._model_name_plot = f"Model {i}: {model._model_name}"
                warnings.warn(
                    f"The _model_name attribute {model._model_name}' is duplicated for models in the `models` you provided. To avoid overlapping model names / plots, the _model_name_plot attribute has been changed to '{model._model_name_plot}'."
                )

        if rename_models is not None:
            model_name_diff = set(rename_models.keys()) - set(all_model_names)
            if model_name_diff:
                warnings.warn(
                    f"""
                    The following model names specified in rename_models are not found in the models:
                    {model_name_diff}
                    """
                )

    return models_list


def _tabulate_etable_md(df, n_coef, n_fixef, n_models, n_model_stats):
    """
    Format and tabulate a DataFrame.

    Parameters
    ----------
    - df (pandas.DataFrame): The DataFrame to be formatted and tabulated.
    - n_coef (int): The number of coefficients.
    - n_fixef (int): The number of fixed effects.
    - n_models (int): The number of models.
    - n_model_stats (int): The number of rows with model statistics.

    Returns
    -------
    - formatted_table (str): The formatted table as a string.
    """
    # Format the DataFrame for tabulate
    table = tabulate(
        df,
        headers="keys",
        showindex=False,
        colalign=["left"] + n_models * ["right"],
    )

    # Split the table into header and body
    header, body = table.split("\n", 1)

    # Add separating line after the third row
    body_lines = body.split("\n")
    body_lines.insert(2, "-" * len(body_lines[0]))
    if n_fixef > 0:
        body_lines.insert(-n_model_stats - n_fixef, "-" * len(body_lines[0]))
    body_lines.insert(-n_model_stats, "-" * len(body_lines[0]))
    body_lines.append("-" * len(body_lines[0]))

    # Join the lines back together
    formatted_table = "\n".join([header, "\n".join(body_lines)])

    # Print the formatted table
    return formatted_table


def _parse_coef_fmt(coef_fmt: str, custom_stats: dict):
    """
    Parse the coef_fmt string.

    Parameters
    ----------
    coef_fmt: str
        The coef_fmt string.
    custom_stats: dict
        A dictionary of custom statistics. Key should be lowercased (e.g., simul_intv).
        If you provide "b", "se", "t", or "p" as a key, it will overwrite the default
        values.

    Returns
    -------
    coef_fmt_elements: str
        The parsed coef_fmt string.
    coef_fmt_title: str
        The title for the coef_fmt string.
    """
    custom_elements = list(custom_stats.keys())
    if any([x in ["b", "se", "t", "p"] for x in custom_elements]):
        raise ValueError(
            "You cannot use 'b', 'se', 't', or 'p' as a key in custom_stats."
        )

    title_map = {
        "b": "Coefficient",
        "se": "Std. Error",
        "t": "t-stats",
        "p": "p-value",
    }

    allowed_elements = [
        "b",
        "se",
        "t",
        "p",
        " ",
        "\n",
        r"\(",
        r"\)",
        r"\[",
        r"\]",
        ",",
        *custom_elements,
    ]
    allowed_elements.sort(key=len, reverse=True)

    coef_fmt_elements = re.findall("|".join(allowed_elements), coef_fmt)
    coef_fmt_title = "".join([title_map.get(x, x) for x in coef_fmt_elements])

    return coef_fmt_elements, coef_fmt_title


def _number_formatter(x: float, **kwargs) -> str:
    """
    Format a number.

    Parameters
    ----------
    x: float
        The series to be formatted.
    digits: int
        The number of digits to round to.
    thousands_sep: bool, optional
        The thousands separator. Default is False.
    scientific_notation: bool, optional
        Whether to use scientific notation. Default is True.
    scientific_notation_threshold: int, optional
        The threshold for using scientific notation. Default is 10_000.
    integer: bool, optional
        Whether to format the number as an integer. Default is False.

    Returns
    -------
    formatted_x: pd.Series
        The formatted series.
    """
    digits = kwargs.get("digits", 3)
    thousands_sep = kwargs.get("thousands_sep", False)
    scientific_notation = kwargs.get("scientific_notation", True)
    scientific_notation_threshold = kwargs.get("scientific_notation_threshold", 10_000)
    integer = kwargs.get("integer", False)

    assert digits >= 0, "digits must be a positive integer"

    if integer:
        digits = 0
    x = np.round(x, digits)

    if scientific_notation and x > scientific_notation_threshold:
        return f"%.{digits}E" % x

    x_str = f"{x:,}" if thousands_sep else str(x)

    if "." not in x_str:
        x_str += ".0"  # Add a decimal point if it's an integer
    _int, _float = str(x_str).split(".")
    _float = _float.ljust(digits, "0")
    return _int if digits == 0 else f"{_int}.{_float}"


def make_table(
    df: pd.DataFrame,
    type: str = "gt",
    notes: str = "",
    rgroup_sep: str = "tb",
    rgroup_display: bool = True,
    caption: Optional[str] = None,
    tab_label: Optional[str] = None,
    texlocation: str = "htbp",
    full_width: bool = False,
    file_name: Optional[str] = None,
    **kwargs,
):
    r"""
    Create a booktab style table in the desired format (gt or tex) from a DataFrame.
    The DataFrame can have a multiindex. Column index used to generate horizonal
    table spanners. Row index used to generate row group names and
    row names. The table can have multiple index levels in columns and up to
    two levels in rows.


    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing the table to be displayed.
    type : str, optional
        Type of table to be created. The default is 'gt'.
    notes : str
        Table notes to be displayed at the bottom of the table.
    rgroup_sep : str
        Whether group names are separated by lines. The default is "tb".
        When output type = 'gt', the options are 'tb', 't', 'b', or '', i.e.
        you can specify whether to have a line above, below, both or none.
        When output type = 'tex' no line will be added between the row groups
        when rgroup_sep is '' and otherwise a line before the group name will be added.
    rgroup_display : bool
        Whether to display row group names. The default is
        True.
    caption : str
        Table caption to be displayed at the top of the table. The default is None.
        When either caption or label is provided the table will be wrapped in a
        table environment.
    tab_label : str
        LaTex label of the table. The default is None. When either caption or label
        is provided the table will be wrapped in a table environment.
    texlocation : str
        Location of the table. The default is 'htbp'.
    full_width : bool
        Whether to expand the table to the full width of the page. The default is False.
    file_name : str
        Name of the file to save the table to. The default is None.
        gt tables will be saved as html files and latex tables as tex files.

    Returns
    -------
    A table in the specified format.
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas DataFrame."
    assert not isinstance(df.index, pd.MultiIndex) or df.index.nlevels <= 2, (
        "Row index can have at most two levels."
    )
    assert type in ["gt", "tex"], "type must be either 'gt' or 'tex'."
    assert rgroup_sep in [
        "tb",
        "t",
        "b",
        "",
    ], "rgroup_sep must be either 'tb', 't', 'b', or ''."
    assert file_name is None or (
        isinstance(file_name, str) and file_name.endswith((".html", ".tex"))
    ), "file_name must end with '.html' or '.tex'."

    # Make a copy of the DataFrame to avoid modifying the original
    dfs = df.copy()

    # Produce LaTeX code if either type is 'tex' or the
    # user has passed a file_name which ends with '.tex'
    if type == "tex" or (isinstance(file_name, str) and file_name.endswith(".tex")):
        # First wrap all cells which contain a line break in a makecell command
        dfs = dfs.map(
            lambda x: f"\\makecell{{{x}}}" if isinstance(x, str) and "\\\\" in x else x
        )
        row_levels = dfs.index.nlevels
        # when the row index has more than one level, we will store
        # the top level to use later to add clines and row group titles
        # and then remove it
        if row_levels > 1:
            # Store the top level of the row index
            top_row_id = dfs.index.get_level_values(0).to_list()
            # Generate a list of the distinct values
            row_groups = list(dict.fromkeys(top_row_id))
            # Generate a list containing the number of rows for each group
            row_groups_len = [top_row_id.count(group) for group in row_groups]
            # Drop the top level of the row index:
            dfs.index = dfs.index.droplevel(0)

        # Style the table
        styler = dfs.style
        # if caption is not None:
        #     styler.set_caption(caption)

        # Generate LaTeX code
        latex_res = styler.to_latex(
            hrules=True,
            multicol_align="c",
            multirow_align="t",
            column_format="l" + "c" * (dfs.shape[1] + dfs.index.nlevels),
        )

        # # Now perform post-processing of the LaTeX code
        # # First split the LaTeX code into lines
        lines = latex_res.splitlines()
        # Find the line number of the \midrule
        line_at = next(i for i, line in enumerate(lines) if "\\midrule" in line)
        # Add space after this \midrule:
        lines.insert(line_at + 1, "\\addlinespace")
        line_at += 1

        # When there are row groups then insert midrules and groupname
        if row_levels > 1 and len(row_groups) > 1:
            # Insert a midrule after each row group
            for i in range(len(row_groups)):
                if rgroup_display:
                    # Insert a line with the row group name & same space around it
                    # lines.insert(line_at+1, "\\addlinespace")
                    lines.insert(line_at + 1, "\\emph{" + row_groups[i] + "} \\\\")
                    lines.insert(line_at + 2, "\\addlinespace")
                    lines.insert(line_at + 3 + row_groups_len[i], "\\addlinespace")
                    line_at += 3
                if (rgroup_sep != "") and (i < len(row_groups) - 1):
                    # For tex output we only either at a line between the row groups or not
                    # And we don't add a line after the last row group
                    line_at += row_groups_len[i] + 1
                    lines.insert(line_at, "\\midrule")
                    lines.insert(line_at + 1, "\\addlinespace")
                    line_at += 1
        else:
            # Add line space before the end of the table
            lines.insert(line_at + dfs.shape[0] + 1, "\\addlinespace")

        # Insert cmidrules (equivalent to column spanners in gt)
        # First find the first line with an occurrence of "multicolumn"
        cmidrule_line_number = None
        for i, line in enumerate(lines):
            if "multicolumn" in line:
                cmidrule_line_number = i + 1
                # Regular expression to find \multicolumn{number}
                pattern = r"\\multicolumn\{(\d+)\}"
                # Find all matches (i.e. values of d) in the LaTeX string & convert to integers
                ncols = [int(match) for match in re.findall(pattern, line)]

                cmidrule_string = ""
                leftcol = 2
                for n in ncols:
                    cmidrule_string += (
                        r"\cmidrule(lr){"
                        + str(leftcol)
                        + "-"
                        + str(leftcol + n - 1)
                        + "} "
                    )
                    leftcol += n
                lines.insert(cmidrule_line_number, cmidrule_string)

        # # Put the lines back together
        latex_res = "\n".join(lines)

        # Wrap in threeparttable to allow for table notes
        if notes is not None:
            latex_res = (
                "\\begin{threeparttable}\n"
                + latex_res
                + "\n\\footnotesize "
                + notes
                + "\n\\end{threeparttable}"
            )
        else:
            latex_res = (
                "\\begin{threeparttable}\n" + latex_res + "\n\\end{threeparttable}"
            )

        # If caption or label specified then wrap in table environment
        if (caption is not None) or (tab_label is not None):
            latex_res = (
                "\\begin{table}["
                + texlocation
                + "]\n"
                + "\\centering\n"
                + ("\\caption{" + caption + "}\n" if caption is not None else "")
                + ("\\label{" + tab_label + "}\n" if tab_label is not None else "")
                + latex_res
                + "\n\\end{table}"
            )

        # Set cell aligment to top
        latex_res = "\\renewcommand\\cellalign{t}\n" + latex_res

        # Set table width to full page width if full_width is True
        # This is done by changing the tabular environment to tabular*
        if full_width:
            latex_res = latex_res.replace(
                "\\begin{tabular}{l", "\\begin{tabularx}{\\linewidth}{X"
            )
            latex_res = latex_res.replace(
                "\\end{tabular}", "\\end{tabularx}\n \\vspace{3pt}"
            )
            # with tabular*
            # latex_res = latex_res.replace("\\begin{tabular}{", "\\begin{tabular*}{\linewidth}{@{\extracolsep{\\fill}}")
            # latex_res = latex_res.replace("\\end{tabular}", "\\end{tabular*}")

        if file_name is not None:
            with open(file_name, "w") as f:
                f.write(latex_res)  # Write the latex code to a file

        if type == "tex":
            return latex_res

    if type == "gt":
        # GT does not support MultiIndex columns, so we need to flatten the columns
        if isinstance(dfs.columns, pd.MultiIndex):
            # Store labels of the last level of the column index (to use as column names)
            col_names = dfs.columns.get_level_values(-1)
            nl = dfs.columns.nlevels
            # As GT does not accept non-unique column names: so to allow for them
            # we just assign column numbers to the lowest index level
            col_numbers = list(map(str, range(len(dfs.columns))))
            # Save the whole column index in order to generate table spanner labels later
            dfcols = dfs.columns.to_list()
            # Then flatten the column index just numbering the columns
            dfs.columns = pd.Index(col_numbers)
            # Store the mapping of column numbers to column names
            col_dict = dict(zip(col_numbers, col_names))
            # Modify the last elements in each tuple in dfcols
            dfcols = [(t[:-1] + (col_numbers[i],)) for i, t in enumerate(dfcols)]
            # And drop the first column as we don't want table spanners on top of the variables
            # WE DON'T NEED THIS WITH ROW INDEX dfcols = dfcols[1:]
        else:
            nl = 1

        rowindex = dfs.index

        # Now reset row index to have the index as columns to be displayed in the table
        dfs.reset_index(inplace=True)

        # And specify the rowname_col and groupname_col
        if isinstance(rowindex, pd.MultiIndex):
            rowname_col = dfs.columns[1]
            groupname_col = dfs.columns[0]
        else:
            rowname_col = dfs.columns[0]
            groupname_col = None

        # Generate the table with GT
        gt = GT(dfs, auto_align=False)

        # When caption is provided, add it to the table
        if caption is not None:
            gt = (
                gt.tab_header(title=caption).tab_options(
                    table_border_top_style="hidden",
                )  # Otherwise line above caption
            )

        if nl > 1:
            # Add column spanners based on multiindex
            # Do this for every level in the multiindex (except the one with the column numbers)
            for i in range(nl - 1):
                col_spanners: dict[str, list[str | int]] = {}
                # Iterate over columns and group them by the labels in the respective level
                for c in dfcols:
                    key = c[i]
                    if key not in col_spanners:
                        col_spanners[key] = []
                    col_spanners[key].append(c[-1])
                for label, columns in col_spanners.items():
                    gt = gt.tab_spanner(label=label, columns=columns, level=nl - 1 - i)
            # Restore column names
            gt = gt.cols_label(**col_dict)

        # Customize the table layout
        gt = (
            gt.tab_source_note(notes)
            .tab_stub(rowname_col=rowname_col, groupname_col=groupname_col)
            .tab_options(
                table_border_bottom_style="hidden",
                stub_border_style="hidden",
                column_labels_border_top_style="solid",
                column_labels_border_top_color="black",
                column_labels_border_bottom_style="solid",
                column_labels_border_bottom_color="black",
                column_labels_border_bottom_width="0.5px",
                column_labels_vlines_color="white",
                column_labels_vlines_width="0px",
                table_body_border_top_style="solid",
                table_body_border_top_width="0.5px",
                table_body_border_top_color="black",
                table_body_hlines_style="none",
                table_body_vlines_color="white",
                table_body_vlines_width="0px",
                table_body_border_bottom_color="black",
                row_group_border_top_style="solid",
                row_group_border_top_width="0.5px",
                row_group_border_top_color="black",
                row_group_border_bottom_style="solid",
                row_group_border_bottom_width="0.5px",
                row_group_border_bottom_color="black",
                row_group_border_left_color="white",
                row_group_border_right_color="white",
                data_row_padding="4px",
                column_labels_padding="4px",
            )
            .cols_align(align="center")
        )

        # Full page width
        if full_width:
            gt = gt.tab_options(table_width="100%")

        # Customize row group display
        if "t" not in rgroup_sep:
            gt = gt.tab_options(row_group_border_top_style="none")
        if "b" not in rgroup_sep:
            gt = gt.tab_options(row_group_border_bottom_style="none")
        if not rgroup_display:
            gt = gt.tab_options(
                row_group_font_size="0px",
                row_group_padding="0px",
            )
        # Save the html code of the table to a file
        if file_name is not None:
            with open(file_name, "w") as f:
                f.write(gt.as_raw_html())

        return gt


def _relabel_index(index, labels=None, stats_labels=None):
    if stats_labels is None:
        if isinstance(index, pd.MultiIndex):
            index = pd.MultiIndex.from_tuples(
                [tuple(labels.get(k, k) for k in i) for i in index]
            )
        else:
            index = [labels.get(k, k) for k in index]
    else:
        # if stats_labels is provided, we relabel the lowest level of the index with it
        if isinstance(index, pd.MultiIndex):
            new_index = []
            for i in index:
                new_index.append(
                    tuple(
                        [labels.get(k, k) for k in i[:-1]]
                        + [stats_labels.get(i[-1], i[-1])]
                    )
                )
            index = pd.MultiIndex.from_tuples(new_index)
        else:
            index = [stats_labels.get(k, k) for k in index]
    return index


def _format_mean_std(
    data: pd.Series, digits: int = 2, newline: bool = True, type=str
) -> str:
    """
    Calculate the mean and standard deviation of a pandas Series and return as a string of the format "mean /n (std)".

    Parameters
    ----------
    data : pd.Series
        The pandas Series for which to calculate the mean and standard deviation.
    digits : int, optional
        The number of decimal places to round the mean and standard deviation to. The default is 2.
    newline : bool, optional
        Whether to add a newline character between the mean and standard deviation. The default is True.
    type : str, optional
        The type of the table output.

    Returns
    -------
    _format_mean_std : str
        The mean and standard deviation of the pandas Series formated as a string.

    """
    mean = data.mean()
    std = data.std()
    if newline:
        if type == "gt":
            return f"{mean:.{digits}f}<br>({std:.{digits}f})"
        elif type == "tex":
            return f"{mean:.{digits}f}\\\\({std:.{digits}f})"
    return f"{mean:.{digits}f} ({std:.{digits}f})"


def dtable(
    df: pd.DataFrame,
    vars: list,
    stats: Optional[list] = None,
    bycol: Optional[list[str]] = None,
    byrow: Optional[str] = None,
    type: str = "gt",
    labels: dict | None = None,
    stats_labels: dict | None = None,
    digits: int = 2,
    notes: str = "",
    counts_row_below: bool = False,
    hide_stats: bool = False,
    **kwargs,
):
    r"""
    Generate descriptive statistics tables and create a booktab style table in
    the desired format (gt or tex).

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing the table to be displayed.
    vars : list
        List of variables to be included in the table.
    stats : list, optional
        List of statistics to be calculated. The default is None, that sets ['count','mean', 'std'].
        All pandas aggregation functions are supported.
    bycol : list, optional
        List of variables to be used to group the data by columns. The default is None.
    byrow : str, optional
        Variable to be used to group the data by rows. The default is None.
    type : str, optional
        Type of table to be created. The default is 'gt'.
        Type can be 'gt' for great_tables, 'tex' for LaTeX or 'df' for dataframe.
    labels : dict, optional
        Dictionary containing the labels for the variables. The default is None.
    stats_labels : dict, optional
        Dictionary containing the labels for the statistics. The default is None.
        The function uses a default labeling which will be replaced by the labels
        in the dictionary.
    digits : int, optional
        Number of decimal places to round the statistics to. The default is 2.
    notes : str
        Table notes to be displayed at the bottom of the table.
    counts_row_below : bool
        Whether to display the number of observations at the bottom of the table.
        Will only be carried out when each var has the same number of obs and when
        byrow is None. The default is False
    hide_stats : bool
        Whether to hide the names of the statistics in the table header. When stats
        are hidden and the user provides no notes string the labels of the stats are
        listed in the table notes. The default is False.
    kwargs : dict
        Additional arguments to be passed to the make_table function.

    Returns
    -------
    A table in the specified format.

    Examples
    --------
    For more examples, take a look at the [regression tables and summary statistics vignette](https://py-econometrics.github.io/pyfixest/table-layout.html).

    ```{python}
    import pyfixest as pf

    # load data
    df = pf.get_data()
    pf.dtable(df, vars = ["Y", "X1", "X2", "f1"])
    ```
    """
    if stats is None:
        stats = ["count", "mean", "std"]
    if labels is None:
        labels = {}
    if stats_labels is None:
        stats_labels = {}
    assert isinstance(df, pd.DataFrame), "df must be a pandas DataFrame."
    assert all(pd.api.types.is_numeric_dtype(df[var]) for var in vars), (
        "Variables must be numerical."
    )
    assert type in ["gt", "tex", "df"], "type must be either 'gt' or 'tex' or 'df'."
    assert byrow is None or byrow in df.columns, (
        "byrow must be a column in the DataFrame."
    )
    assert bycol is None or all(col in df.columns for col in bycol), (
        "bycol must be a list of columns in the DataFrame."
    )

    # Default stats labels dictionary
    stats_dict = {
        "count": "N",
        "mean": "Mean",
        "std": "Std. Dev.",
        "mean_std": "Mean (Std. Dev.)",
        "mean_newline_std": "Mean (Std. Dev.)",
        "min": "Min",
        "max": "Max",
        "var": "Variance",
        "median": "Median",
    }
    stats_dict.update(stats_labels or {})

    # Define custom aggregation functions
    def mean_std(x):
        return _format_mean_std(x, digits=digits, newline=False, type=type)

    def mean_newline_std(x):
        return _format_mean_std(x, digits=digits, newline=True, type=type)

    # Create a dictionary to map stat names to custom functions
    custom_funcs = {"mean_std": mean_std, "mean_newline_std": mean_newline_std}

    # Prepare the aggregation dictionary allowing custom functions
    agg_funcs = {var: [custom_funcs.get(stat, stat) for stat in stats] for var in vars}

    # Calculate the desired statistics
    if (byrow is not None) and (bycol is not None):
        bylist = [byrow, *bycol]
        res = df.groupby(bylist).agg(agg_funcs)
    if (byrow is None) and (bycol is None):
        res = df.agg(agg_funcs)
    elif (byrow is not None) and (bycol is None):
        res = df.groupby(byrow).agg(agg_funcs)
    elif (byrow is None) and (bycol is not None):
        res = df.groupby(bycol).agg(agg_funcs)

    # Set counts_row_below to false when byrow is not None
    # or when 'count' is not in stats
    if (byrow is not None) or ("count" not in stats):
        counts_row_below = False

    # Round all floats to required decimal places
    # Convert to string to preserve the formatting
    format_string = ",." + str(digits) + "f"

    # Reshaping of table (just transpose when no multiindex)
    if res.columns.nlevels == 1:
        # Check whether number of obs should be displayed at the bottom
        if counts_row_below:
            # Only when all counts are the same within each row
            if res.loc["count"].nunique() == 1:
                # collect the number of obs
                nobs = res.loc["count"].iloc[0]
                # Drop the count row
                res = res.drop("count", axis=0)
                if "count" in stats:
                    stats.remove("count")
            else:
                counts_row_below = False

        # Transpose
        res = res.transpose(copy=True)

        # print(res)
        # Format the statistics
        for col in res.columns:
            # Format the statistics
            # for some reason count stats are displayed as floats when no multiindex,
            # so we need to convert them to integers
            if res[col].name == "count":
                res[col] = res[col].apply(lambda x: f"{x:.0f}")
            elif res[col].dtype == float:
                res[col] = res[col].apply(lambda x: f"{x:{format_string}}")

        # Add the number of observations at the bottom of the table
        if counts_row_below:
            obs_row = [str(int(nobs))] + [""] * (len(res.columns) - 1)
            res.loc[stats_dict["count"]] = obs_row

    else:
        # When there is a multiindex in the columns
        # First check whether number of obs should be displayed at the bottom
        if counts_row_below:
            # collect the number of obs for each row
            count_columns = res.xs("count", axis=1, level=-1)
            # Ensure count_columns is always a DataFrame
            if isinstance(count_columns, pd.Series):
                count_columns = count_columns.to_frame()
            # when all counts are the same within each row,
            # generate a vector with the counts
            if count_columns.nunique(axis=1).eq(1).all():
                nobs = count_columns.iloc[:, 0]
                # Drop the count column
                res = res.drop("count", axis=1, level=-1)
                if "count" in stats:
                    stats.remove("count")
                # And append the counts as an additional column
                # with the value being assigned to the column of the first stat
                # and labeled as defined in the stats_dict
                res[stats_dict["count"], stats[0]] = nobs
            else:
                counts_row_below = False

        # Format the statistics
        for col in res.columns:
            if res[col].dtype == float:
                res[col] = res[col].apply(lambda x: f"{x:{format_string}}")

        # Now some reshaping to bring the multiindex dataframe in the form of a typical descriptive statistics table
        res = pd.DataFrame(res.stack(level=0, future_stack=True))

        # First bring the variables to the rows:
        # Assign name to the column index
        res.columns.names = ["Statistics"]
        if bycol is not None:
            # Then bring the column objects to the columns:
            res = pd.DataFrame(res.unstack(level=tuple(bycol)))
            # Finally we want to have the objects first and then the statistics
            if not isinstance(res.columns, pd.MultiIndex):
                res.columns = pd.MultiIndex.from_tuples(res.columns)  # type: ignore
            res.columns = res.columns.reorder_levels([*bycol, "Statistics"])
            # And sort it properly by the variables
            # (we want to preserve the order of the lowest level for the stats)
            levels_to_sort = list(range(res.columns.nlevels - 1))
            res = res.sort_index(axis=1, level=levels_to_sort, sort_remaining=False)

        # When hide_stats is True, we remove the names of the statistics
        # And add a note to the table listing the statistics when the user
        # has not provided a notes string
        if hide_stats:
            res.columns = res.columns.droplevel(-1)
            if notes == "":
                notes = (
                    "Note: Displayed statistics are "
                    + ", ".join([stats_dict.get(k, k) for k in stats])
                    + "."
                )

    # Replace all NaNs with empty strings
    res = res.fillna("")

    # Relabel Variable names in row and column indices
    res.columns = _relabel_index(res.columns, labels, stats_dict)
    res.index = _relabel_index(res.index, labels)

    # When counts_row_below: Turn row index into a multiindex
    # to set up a second panel for the number of observations
    # that make_table will thus separate by a line
    if counts_row_below:
        res.index = pd.MultiIndex.from_tuples([("stats", i) for i in res.index])
        # Modify the last tuple in the MultiIndex
        new_index = list(res.index)
        new_index[-1] = ("nobs", stats_dict["count"])
        res.index = pd.MultiIndex.from_tuples(new_index)

    # Show row groups iff byrow is not None
    rgroup_display = byrow is not None

    # Generate the table
    if type in ["gt", "tex"]:
        # And make a booktab
        return make_table(
            res, type=type, notes=notes, rgroup_display=rgroup_display, **kwargs
        )
    else:
        return res



import numpy as np
import pytest

from pyfixest.estimation.solvers import solve_ols


def test_solve_ols_simple_2x2():
    # Test case 1: Simple 2x2 system
    tZX = np.array([[4, 2], [2, 3]])
    tZY = np.array([10, 8])
    solver = "scipy.linalg.solve"
    solution = solve_ols(tZX, tZY, solver)
    assert np.allclose(solution, np.array([1.75, 1.5]))
    # Verify solution satisfies the system
    assert np.allclose(tZX @ solution, tZY)


def test_solve_ols_identity():
    # Test case 2: Identity matrix
    tZX = np.eye(2)
    tZY = np.array([1, 2])
    solver = "scipy.linalg.solve"
    assert np.allclose(solve_ols(tZX, tZY, solver), tZY)


@pytest.mark.parametrize(
    argnames="solver",
    argvalues=[
        "scipy.linalg.solve",
        "np.linalg.lstsq",
        "np.linalg.solve",
        "scipy.sparse.linalg.lsqr",
        "jax",
    ],
    ids=[
        "scipy.linalg.solve",
        "np.linalg.lstsq",
        "np.linalg.solve",
        "scipy.sparse.linalg.lsqr",
        "jax",
    ],
)
def test_solve_ols_different_solvers(solver):
    # Test case 3: Test different solvers give same result
    tZX = np.array([[4, 2], [2, 3]])
    tZY = np.array([10, 8])
    solution = solve_ols(tZX, tZY, solver)
    assert np.allclose(solution, np.array([1.75, 1.5]))
    # Verify solution satisfies the system
    assert np.allclose(tZX @ solution, tZY)


def test_solve_ols_invalid_solver():
    # Test case 4: Invalid solver
    tZX = np.array([[1, 2], [3, 4]])
    tZY = np.array([5, 6])
    with pytest.raises(ValueError):
        solve_ols(tZX, tZY, "invalid_solver")



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from pyfixest.core import count_fixef_fully_nested_all

# Import your module here - adjust import path as needed
from pyfixest.estimation.numba.nested_fixef_nb import (
    _count_fixef_fully_nested_all as count_fixef_fully_nested_all_nb,
)


@pytest.mark.parametrize(
    "fn", [count_fixef_fully_nested_all, count_fixef_fully_nested_all_nb]
)
def test_count_fixef_fully_nested_basic(benchmark, fn):
    """Basic test for count_fixef_fully_nested_all_rs function."""
    # Setup test data
    all_fe = np.array(["fe1", "fe2", "cluster1"])
    cluster_names = np.array(["cluster1"])

    # Fixed effects data where fe1 is nested in cluster1
    fe_data = np.array(
        [
            [1, 5, 0],  # row 1 - fe1=1, fe2=5, cluster1=0
            [1, 6, 1],  # row 2 - fe1=1, fe2=6, cluster1=1
            [2, 5, 0],  # row 3 - fe1=2, fe2=5, cluster1=0
            [2, 6, 1],  # row 4 - fe1=2, fe2=6, cluster1=1
        ],
        dtype=np.uintp,
    )

    # Cluster data - one column for the "cluster1" variable
    cluster_data = np.array(
        [
            [10],  # fe1=1 always maps to cluster=10
            [10],
            [20],  # fe1=2 always maps to cluster=20
            [20],
        ],
        dtype=np.uintp,
    )

    # Call function
    mask, count = benchmark(fn, all_fe, cluster_names, cluster_data, fe_data)

    # Expected results
    expected_mask = np.array(
        [True, False, True]
    )  # fe1 is nested, fe2 is not, cluster1 is a cluster
    expected_count = 2

    # Assertions
    assert_array_equal(np.array(mask), expected_mask)
    assert count == expected_count



import warnings
from collections.abc import Mapping
from functools import partial
from typing import Any, Callable, Literal, Optional, Union, cast

import numba as nb
import numpy as np
import pandas as pd
from scipy.linalg import cho_factor, solve_triangular
from scipy.stats import norm

from pyfixest.estimation.feols_ import Feols
from pyfixest.estimation.FormulaParser import FixestFormula
from pyfixest.estimation.literals import (
    QuantregMethodOptions,
    SolverOptions,
)
from pyfixest.estimation.quantreg.frisch_newton_ip import (
    frisch_newton_solver,
)
from pyfixest.estimation.quantreg.utils import get_hall_sheather_bandwidth
from pyfixest.estimation.vcov_utils import bucket_argsort


class Quantreg(Feols):
    "Quantile regression model."

    def __init__(
        self,
        FixestFormula: FixestFormula,
        data: pd.DataFrame,
        ssc_dict: dict[str, Union[str, bool]],
        drop_singletons: bool,
        drop_intercept: bool,
        weights: Optional[str],
        weights_type: Optional[str],
        collin_tol: float,
        fixef_tol: float,
        fixef_maxiter: int,
        lookup_demeaned_data: dict[str, pd.DataFrame],
        solver: SolverOptions = "np.linalg.solve",
        demeaner_backend: Literal["numba", "jax"] = "numba",
        store_data: bool = True,
        copy_data: bool = True,
        lean: bool = False,
        context: Union[int, Mapping[str, Any]] = 0,
        sample_split_var: Optional[str] = None,
        sample_split_value: Optional[Union[str, int]] = None,
        quantile: float = 0.5,
        method: QuantregMethodOptions = "fn",
        quantile_tol: float = 1e-06,
        quantile_maxiter: Optional[int] = None,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__(
            FixestFormula=FixestFormula,
            data=data,
            ssc_dict=ssc_dict,
            drop_singletons=drop_singletons,
            drop_intercept=drop_intercept,
            weights=weights,
            weights_type=weights_type,
            collin_tol=collin_tol,
            fixef_tol=fixef_tol,
            fixef_maxiter=fixef_maxiter,
            lookup_demeaned_data=lookup_demeaned_data,
            solver=solver,
            store_data=store_data,
            copy_data=copy_data,
            lean=lean,
            sample_split_var=sample_split_var,
            sample_split_value=sample_split_value,
            context=context,
            demeaner_backend=demeaner_backend,
        )

        warnings.warn(
            """
           The Quantile Regression implementation is experimental and may change in future releases.
           But mostly, we expect the API to remain unchanged.
           """,
            FutureWarning,
        )

        self._supports_wildboottest = False
        self._support_crv3_inference = False
        self._supports_cluster_causal_variance = False

        self._quantile = quantile
        self._method = f"quantreg_{method}"
        self._quantile_tol = quantile_tol
        self._quantile_maxiter = quantile_maxiter

        self._model_name = (
            FixestFormula.fml
            if self._sample_split_var is None
            else f"{FixestFormula.fml} (Sample: {self._sample_split_var} = {self._sample_split_value})"
        )
        # update with quantile name
        self._model_name = f"{self._model_name} (q = {quantile})"
        self._model_name_plot = self._model_name

        self._seed = seed

        # later set in fit method, consant for different quantiles q -> can be reused
        self._chol = None
        self._P = None

        self._method_map: dict[
            str,
            Callable[
                ...,
                tuple[
                    np.ndarray,
                    bool,
                    int,
                    np.ndarray,
                    np.ndarray,
                    np.ndarray,
                    np.ndarray,
                    np.ndarray,
                ],
            ],
        ] = {
            "fn": partial(
                self.fit_qreg_fn,
                q=self._quantile,
                tol=self._quantile_tol,
                maxiter=self._quantile_maxiter,
                beta_init=None,
            ),
            "pfn": partial(
                self.fit_qreg_pfn,
                q=self._quantile,
                rng=np.random.default_rng(self._seed),
                tol=self._quantile_tol,
                maxiter=self._quantile_maxiter,
                beta_init=None,
            ),
        }

        try:
            self._fit = self._method_map[method]
        except KeyError as exc:
            valid = ", ".join(self._method_map)
            raise ValueError(f"`method` must be one of {{{valid}}}") from exc

    def to_array(self):
        "Turn estimation DataFrames to np arrays."
        self._Y, self._X, self._Z = (
            self._Y.to_numpy(),
            self._X.to_numpy(),
            self._X.to_numpy(),
        )

    def prepare_model_matrix(self):
        "Prepare model inputs for estimation."
        super().prepare_model_matrix()

        if self._fe is not None:
            raise NotImplementedError(
                "Fixed effects are not yet supported for Quantile Regression."
            )

    def get_fit(self) -> None:
        """Fit a quantile regression model using the interior point method."""
        res = self._fit(X=self._X, Y=self._Y)

        self._beta_hat = res[0]
        self._has_converged = res[1]
        self._it = res[2]
        self._x_final = res[3]
        self._s_final = res[4]
        self._z_final = res[5]
        self._w_final = res[6]
        self._y_final = res[7]

        self._Y_hat_link = self._X @ self._beta_hat
        self._Y_hat_response = self._Y_hat_link

        self._u_hat = self._Y.flatten() - self._X @ self._beta_hat
        self._hessian = self._X.T @ self._X
        self._bread = np.linalg.inv(self._hessian)

    def fit_qreg_fn(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        q: float,
        tol: Optional[float] = None,
        maxiter: Optional[int] = None,
        beta_init: Optional[np.ndarray] = None,
    ) -> tuple[
        np.ndarray,
        bool,
        int,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]:
        """Fit a quantile regression model using the Frisch-Newton Interior Point Solver."""
        N, k = X.shape
        if tol is None:
            tol = 1e-06
        if maxiter is None:
            maxiter = N

        # compute cholesky once outside of FN loop
        # if self._chol is None or self._P is None:
        _chol, _ = cho_factor(X.T @ X, lower=True, check_finite=False)
        _chol = np.atleast_2d(_chol)
        _P = solve_triangular(_chol, X.T, lower=True, check_finite=False)
        # if self._chol is None or self._P is None:
        #    raise ValueError("...")

        fn_res = frisch_newton_solver(
            A=X.T,
            b=(1 - q) * X.T @ np.ones(N),
            c=-Y,
            u=np.ones(N),
            q=q,
            tol=tol,
            max_iter=maxiter,
            backoff=0.9995,
            beta_init=beta_init,
            chol=cast(np.ndarray, _chol),
            P=cast(np.ndarray, _P),
        )

        has_converged = fn_res[1]
        it = fn_res[2]

        if not has_converged:
            warnings.warn(
                f"The Frisch-Newton Interior Point solver has not converged after {it} iterations."
            )

        return fn_res

    def fit_qreg_pfn(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        q: float,
        m: Optional[float] = None,
        tol: Optional[float] = None,
        maxiter: Optional[int] = None,
        beta_init: Optional[np.ndarray] = None,
        rng: Optional[np.random.Generator] = None,
        eta: Optional[float] = None,
    ) -> tuple[
        np.ndarray,
        bool,
        int,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
        np.ndarray,
    ]:
        """Fit a quantile regression model using the Frisch-Newton Interior Point Solver with pre-processing."""
        N, k = X.shape
        if tol is None:
            tol = 1e-06
        if maxiter is None:
            maxiter = N
        if rng is None:
            rng = np.random.default_rng()
        if beta_init is None:
            beta_init = np.zeros(k)
        if m is None:
            m = 0.8
        if eta is None:
            eta = 2 / 3

        max_bad_fixups = 3
        n_bad_fixups = 0

        has_converged = False
        compute_beta_init = beta_init is None

        # m constant should be set by user
        m = 0.8
        n_init = int(np.ceil((k * N) ** (eta)))
        M = int(np.maximum(N, np.ceil(m * n_init)))

        while not has_converged:
            if compute_beta_init:
                # get initial sample
                idx_init = rng.choice(N, size=n_init, replace=False)
                beta_hat_init = self.fit_qreg_fn(
                    X[idx_init, :], Y[idx_init], q=q, tol=tol, maxiter=maxiter
                )[0]

            else:
                beta_hat_init = beta_init

            r_init = Y.flatten() - X @ beta_hat_init
            # conservative estimate of sigma^2
            z = np.sqrt(np.dot(r_init, r_init) / (N - k))
            rz = r_init / z

            ql = max(0, q - M / (2 * N))
            qu = min(1, q + M / (2 * N))

            JL = rz < np.quantile(rz, ql)
            JH = rz > np.quantile(rz, qu)

            while not has_converged and n_bad_fixups < max_bad_fixups:
                keep = ~(JL | JH)
                X_sub = X[keep, :]
                Y_sub = Y[keep, :]

                if np.any(JL):
                    X_neg = np.sum(X[JL, :], axis=0)
                    Y_neg = np.sum(Y[JL])
                    X_sub = np.concatenate([X_sub, X_neg.reshape((1, self._k))], axis=0)
                    Y_sub = np.concatenate([Y_sub, Y_neg.reshape((1, 1))], axis=0)
                if np.any(JH):
                    X_pos = np.sum(X[JH, :], axis=0)
                    Y_pos = np.sum(Y[JH])
                    X_sub = np.concatenate([X_sub, X_pos.reshape(1, self._k)], axis=0)
                    Y_sub = np.concatenate([Y_sub, Y_pos.reshape((1, 1))], axis=0)

                # solve the modified problem
                fn_res = self.fit_qreg_fn(X=X_sub, Y=Y_sub, q=q)
                beta_hat = fn_res[0]

                r = Y.flatten() - X @ beta_hat

                # count wrong predictions and get their indices
                mis_L = JL & (r > 0)
                mis_H = JH & (r < 0)
                n_bad = np.sum(mis_L) + np.sum(mis_H)

                if n_bad == 0:
                    has_converged = True
                    break
                elif n_bad > 0.1 * M:
                    warnings.warn("Too many bad fixups. Doubling m.")
                    n_init = min(N, 2 * n_init)
                    M = int(np.ceil(m * n_init))
                    n_bad_fixups += 1
                    compute_beta_init = True
                    break

                else:
                    JL = JL & ~mis_L
                    JH = JH & ~mis_H

        if not has_converged:
            warnings.warn(
                "The Frisch-Newton Interior Point solver with preprocessing has not converged after 3 bad fixups."
            )

        return fn_res

    def _vcov_iid(self):
        "Implement the kernel-based sandwich estimator from Powell (1991)."
        q = self._quantile
        N = self._N
        X = self._X
        Y = self._Y
        u_hat = self._u_hat

        h = get_hall_sheather_bandwidth(q=q, N=N)
        # interquartile range of u_hat - this is what both quantreg and statsmodels use
        # (all three logical lines below in fact)
        rq = np.quantile(np.abs(u_hat), 0.75) - np.quantile(np.abs(u_hat), 0.25)
        sigma = np.std(Y)
        hk = np.minimum(sigma, rq / 1.34) * (norm.ppf(q + h) - norm.ppf(q - h))

        # uniform kernel
        f = 1 / (2 * N * hk) * np.sum(np.abs(u_hat) < hk)

        D = X.T @ X
        Dinv = np.linalg.inv(D)

        return 1 / (f**2) * q * (1 - q) * Dinv

    def _vcov_hetero(self):
        "Implement the kernel-based sandwich estimator from Powell (1991) for heteroskedasticity robust inference."
        q = self._quantile
        N = self._N
        X = self._X
        Y = self._Y
        u_hat = self._u_hat

        h = get_hall_sheather_bandwidth(q=q, N=N)
        # interquartile range of u_hat
        rq = np.quantile(np.abs(u_hat), 0.75) - np.quantile(np.abs(u_hat), 0.25)
        sigma = np.std(Y)
        hk = np.minimum(sigma, rq / 1.34) * (norm.ppf(q + h) - norm.ppf(q - h))

        # uniform kernel
        f = 1 / (2 * N * hk) * np.sum(np.abs(u_hat) < hk)

        D = X.T @ X
        C = f * D
        Cinv = np.linalg.inv(C)

        return q * (1 - q) * Cinv @ D @ Cinv

    def _vcov_nid(self) -> np.ndarray:
        """
        Compute nonparametric IID (NID) vcov matrix using the Hall-Sheather bandwidth
        as developed in Hendricks and Koenker (1991).
        Note: the estimator is actually heteroskedasticity robust, despite its name.
        'nid' stands for 'non-iid'.
        For details, see page 80 in Koenker's "Quantile Regression" (2005) book.
        """
        q = self._quantile
        N = self._N
        X = self._X

        h = get_hall_sheather_bandwidth(q=q, N=N)

        beta_hat_plus = self._fit(
            X=self._X,
            Y=self._Y,
            q=self._quantile + h,
            beta_init=self._beta_hat if self._method == "pfn" else None,
        )[0]
        beta_hat_minus = self._fit(
            X=self._X,
            Y=self._Y,
            q=self._quantile - h,
            beta_init=self._beta_hat if self._method == "pfn" else None,
        )[0]

        # eps: small tolerance parameter to avoid division by zero
        # when di = 0; set to sqrt of machine epsilon in quantreg
        eps = np.finfo(float).eps ** 0.5
        # equation (2)
        di = X @ (beta_hat_plus - beta_hat_minus)
        # equation (3)
        Fplus = np.maximum(0, (2 * h) / (di - eps))

        # general Huber structure, see page 74 in Koenker.
        J = X.T @ X
        XFplus = X * np.sqrt(Fplus[:, np.newaxis])
        H = XFplus.T @ XFplus
        Hinv = np.linalg.inv(H)

        return q * (1 - q) * Hinv @ J @ Hinv

    def _vcov_crv1(self, clustid: np.ndarray, cluster_col: np.ndarray):
        """
        Implement cluster robust variance estimator for quantile regression following
        Parente and Santos Silva, 2016.
        """
        if len(self._clustervar) > 1:
            raise NotImplementedError(
                "Multiway clustering is not (yet) supported for quantile regression."
            )

        X = self._X
        N, _ = X.shape
        q = self._quantile
        u_hat = self._u_hat

        # kappa: median absolute deviation of the a-th quantile regression residuals
        kappa = np.median(np.abs(u_hat - np.median(u_hat)))
        h_G = get_hall_sheather_bandwidth(q=q, N=N)
        delta = kappa * (norm.ppf(q + h_G) - norm.ppf(q - h_G))

        vcov = _crv1_vcov_loop(
            X=X, clustid=clustid, cluster_col=cluster_col, q=q, u_hat=u_hat, delta=delta
        )

        return vcov

    @property
    def objective_value(self):
        "Compute the total loss of the quantile regression model."
        return np.sum(np.abs(self._u_hat) * (self._quantile - (self._u_hat < 0)))

    def get_performance(self):
        "Compute performance metrics for the quantile regression model."
        # self._pseudo_r2 = 1 -
        pass
        # self.objective_value


@nb.njit(parallel=False)
def _crv1_vcov_loop(
    X: np.ndarray,
    clustid: np.ndarray,
    cluster_col: np.ndarray,
    q: float,
    u_hat: np.ndarray,
    delta: float,
) -> np.ndarray:
    _, k = X.shape

    A = np.zeros((k, k))
    B = np.zeros((k, k))
    g_indices, g_locs = bucket_argsort(cluster_col)

    eps = 1e-7

    for g in clustid:
        start = g_locs[g]
        end = g_locs[g + 1]
        g_index = g_indices[start:end]

        Xg = X[g_index, :]
        ug = u_hat[g_index]

        ng = g_index.size
        for i in range(ng):
            Xgi = Xg[i, :]
            psi_i = q - 1.0 * (ug[i] <= eps)
            for j in range(ng):
                Xgj = Xg[j, :]
                psi_j = q - 1.0 * (ug[j] <= eps)
                A += np.outer(Xgi, Xgj) * psi_i * psi_j

            mask_i = (np.abs(ug[i]) < delta) * 1.0
            B += np.outer(Xgi, Xgi) * mask_i

    B /= 2 * delta

    return np.linalg.inv(B) @ A @ np.linalg.inv(B)



import re
from itertools import product
from typing import Optional, Union

from pyfixest.errors import (
    DuplicateKeyError,
    EndogVarsAsCovarsError,
    InstrumentsAsCovarsError,
    UnderDeterminedIVError,
    UnsupportedMultipleEstimationSyntax,
)


class FixestFormulaParser:
    """
    A class for deparsing formulas with multiple estimation syntax.

    Methods
    -------
        __init__(self, fml): Constructor method that initializes the object with
        a given formula.
        get_fml_dict(self): Returns a dictionary of all fevars & formula without
        fevars.
        get_var_dict(self): Returns a dictionary of all fevars and list of covars
        and depvars used in regression with those fevars.

    """

    def __init__(self, fml: str):
        """
        Initialize the object with a given formula using constructor method.

        Parameters
        ----------
        fml : str
            A one-to three sided formula string in the form
            "Y1 + Y2 ~ X1 + X2 | FE1 + FE2 | endogvar ~ exogvar".

        Returns
        -------
            None

        """
        depvars, covars, fevars, endogvars, instruments = _deparse_fml(fml)

        # Parse all individual formula components that allow for
        # multiple estimations into dictionaries that separate a 'constant'
        # part common to all estimations from a varying part with key of the
        # 'type' of variation, e.g. 'sw' or 'csw'.
        depvars_list = depvars.split("+")
        covars_dict = _input_formula_to_dict(
            covars
        )  # e.g. {'constant': [], 'csw': ['X1', 'X2']}
        fevars_dict = _input_formula_to_dict(fevars)  # e.g. {'constant': ['f1^f2']}
        # Now parse all formula components in covars_dict, fevars_dict into lists
        # of formulas that can be used in the estimation.
        # E.g. {'constant': [], 'csw': ['X1', 'X2']} becomes ['X1', 'X1+X2']
        # and {'constant': ['f1^f2']} becomes ['f1^f2'].
        covars_formulas_list = _dict_to_list_of_formulas(
            covars_dict
        )  # evaluate self.covars to list: ['X1', 'X1+X2']
        fevars_formula_list = _dict_to_list_of_formulas(fevars_dict)  # ['f1^f2']

        self.condensed_fml_dict = collect_fml_dict(
            fevars_formula_list, depvars_list, covars_formulas_list
        )

        # now repeat for IV:
        self.is_iv = False
        self.condensed_fml_dict_iv = None
        endogvars_list = []
        instruments_formulas_list = []
        if endogvars is not None and instruments is not None:
            self.is_iv = True
            endogvars_list = endogvars.split("+")
            instruments_dict = _input_formula_to_dict(instruments)
            instruments_formulas_list = _dict_to_list_of_formulas(instruments_dict)
            self.condensed_fml_dict_iv = collect_fml_dict(
                fevars_formula_list, endogvars_list, instruments_formulas_list
            )

        self.FixestFormulaDict: dict[str, list[FixestFormula]] = {}
        self.populate_fixest_formula_dict(
            depvars_list=depvars_list,
            covars_formulas_list=covars_formulas_list,
            fevars_formula_list=fevars_formula_list,
            endogvars_list=endogvars_list,
            instruments_formulas_list=instruments_formulas_list,
        )

    def add_to_FixestFormulaDict(
        self,
        depvar: str,
        covar: str,
        fval: str,
        endogvar: Optional[str] = None,
        instrument: Optional[str] = None,
    ):
        """
        Add a FixestFormula object to the FixestFormulaDict.

        This method initializes a FixestFormula object with the specified
        dependent variable, covariates, fixed effect variable, optional endogenous
        variable, and optional instrument. It then generates both first and
        second stage formulas, as well as the combined formula,
        and adds the FixestFormula object to the FixestFormulaDict
        under the key specified by the `fval` parameter.
        If the `fval` key does not exist in the dictionary, it is created.

        Parameters
        ----------
        depvar : str
            The name of the dependent variable.
        covar : str
            The covariates included in the model, formatted as a string. E.g. X1+X2.
        fval : str
            The fixed effect variable included in the model.
        endogvar : Optional[str], optional
            The endogenous variable, if applicable.
        instrument : Optional[str], optional
            The instrument for the endogenous variable, if applicable.

        Notes
        -----
        - This method does not return any value.
        - It updates the `FixestFormulaDict` attribute of the instance by
          appending a new FixestFormula object or creating a new list for
          a previously unused `fval`.
        """
        FixestFML = FixestFormula(
            depvar=depvar,
            covar=covar,
            fval=fval,
            endogvars=endogvar,
            instruments=instrument,
        )
        FixestFML.get_first_and_second_stage_fml()
        FixestFML.get_fml()

        if fval not in self.FixestFormulaDict:
            self.FixestFormulaDict[fval] = []
        self.FixestFormulaDict[fval].append(FixestFML)

    def populate_fixest_formula_dict(
        self,
        depvars_list: list[str],
        covars_formulas_list: list[str],
        fevars_formula_list: list[str],
        endogvars_list: list[str],
        instruments_formulas_list: list[str],
    ):
        """
        Populate the FixestFormulaDict with FixestFormula objects.

        Iterates through all combinations of provided lists of dependent
        variables, covariate formulas, fixed effect variables, endogenous
        variables (if applicable), and instrument formulas (if applicable)
        to populate the `FixestFormulaDict` with corresponding FixestFormula
        objects. This method supports models with and without instrumental
        variables (IV).

        Parameters
        ----------
        depvars_list : list
            A list of dependent variable names.
        covars_formulas_list : list
            A list of covariate formulas.
        fevars_formula_list : list
            A list of fixed effect variable formulas.
        endogvars_list : list
            A list of endogenous variables, used if the model is an IV model.
        instruments_formulas_list : list
            A list of instrument formulas, used if the model is an IV model.

        Notes
        -----
        - This method does not return any value.
        - It conditionally processes combinations based on whether the model is
          identified as an IV model, indicated by the `is_iv` attribute of the
          instance.
        - For IV models, it combines variables from `depvars_list`,
          `covars_formulas_list`, `fevars_formula_list`,
          `endogvars_list`, and `instruments_formulas_list`.
        - For non-IV models, it combines variables from `depvars_list`,
          `covars_formulas_list`, and `fevars_formula_list` only.
        """
        if self.is_iv:
            for depvar, covar, fval, endogvar, instrument in product(
                depvars_list,
                covars_formulas_list,
                fevars_formula_list,
                endogvars_list,
                instruments_formulas_list,
            ):
                self.add_to_FixestFormulaDict(depvar, covar, fval, endogvar, instrument)
        else:
            for depvar, covar, fval in product(
                depvars_list, covars_formulas_list, fevars_formula_list
            ):
                self.add_to_FixestFormulaDict(depvar, covar, fval)

    def set_fixest_multi_flag(self):
        """
        Set a flag to indicate whether multiple estimations are being performed or not.

        Simple check if `all_fitted_models` has length greater than 1.
        Throws an error if multiple estimations are being performed with IV estimation.
        Args:
            None
        Returns:
            None
        """
        if (
            len(self.FixestFormulaDict) == 1
            and isinstance(next(iter(self.FixestFormulaDict.values())), list)
            and len(next(iter(self.FixestFormulaDict.values()))) == 1
        ):
            self._is_multiple_estimation = False
        else:
            self._is_multiple_estimation = True
            if self.is_iv:
                raise NotImplementedError(
                    """
                    Multiple Estimations is currently not supported with IV.
                    This is mostly due to insufficient testing and will be possible
                    with a future release of PyFixest.
                    """
                )


class FixestFormula:
    """
    A class with information contained in model formulas.

    Attributes
    ----------
    _depvar : str
        The dependent variable in the model.
    _covar : str
        The covariates in the model, separated by '+'.
    _fval : str
        An optional fixed effect variable included in the model.
        Separated by "+". "0" if no fixed effect in the model.
    _endogvars : str, optional
        Endogenous variables in the model, separated by '+'.
    _instruments : str, optional
        Instrumental variables for the endogenous variables, separated by '+'.

    Methods
    -------
    get_first_and_second_stage_fml(self):
        Assigns first and second stage formulas for IV models
        to instance variables.

    get_fml(self):
        Constructs and stores a formula based on instance attributes.

    check_syntax(self):
        Validates the formula by checking if instruments are incorrectly
        specified as covariates.

    Raises
    ------
    InstrumentsAsCovarsError
        If any instrumental variables are also specified as covariates
        in the model formula.

    """

    def __init__(
        self,
        depvar: str,
        covar: str,
        fval: Optional[str] = None,
        endogvars: Optional[str] = None,
        instruments: Optional[str] = None,
    ):
        self._depvar = depvar
        self._covar = covar
        if fval is None:
            fval = "0"
        self._fval = fval
        self._endogvars = endogvars
        self._instruments = instruments

    def get_first_and_second_stage_fml(self):
        """Assign first and second stage formulas."""
        self.fml_second_stage, self.fml_first_stage = _get_first_and_second_stage_fml(
            depvar=self._depvar,
            covar=self._covar,
            fval=self._fval,
            endogvar=self._endogvars,
            instruments=self._instruments,
        )

    def get_fml(self):
        """
        Construct and stores a Wilkinson formula..

        This method combines dependent variable, covariates, endogenous variables,
        instrumental variables, and an optional fixed value to construct a statistical
        model formula. This formula is then stored in the instance's `fml` attribute.
        The general structure of the formula is
        `depvar ~ covar | fval | endogvars ~ instruments`.
        Spaces in the formula are removed before storing.

        Attributes Used
        ---------------
        _depvar : str
            The dependent variable in the model.
        _covar : str
            The covariates in the model, separated by '+'.
        _fval : str
            An optional fixed value to be included in the model.
            If set to "0", it is ignored.
        _endogvars : str, optional
            Endogenous variables in the model, separated by '+'.
            If `None`, this part of the formula is omitted.
        _instruments : str, optional
            Instrumental variables in the model, separated by '+'.
            Relevant only if `_endogvars` is not `None`.

        Notes
        -----
        - The method does not return any value but updates the instance's `fml`
          attribute with the constructed formula.
        - This method assumes that `_depvar`, `_covar`, and `_fval` are always
          provided and treats `_endogvars` and `_instruments` as optional components
          of the formula.
        """
        depvar = self._depvar
        covar = self._covar
        fval = self._fval
        endogvars = self._endogvars
        instruments = self._instruments

        fml = f"{depvar} ~ {covar}"
        fml_iv = f"| {endogvars} ~ {instruments}" if endogvars is not None else None

        fml_fval = f"| {fval}" if fval != "0" else None

        if fml_iv is not None:
            fml += fml_iv

        if fml_fval is not None:
            fml += fml_fval

        self.fml = fml.replace(" ", "")

    def check_syntax(self):
        """
        Check if any instrument variables are mistakenly specified as covariates.

        This method processes the instrument and covariate strings stored in
        the instance's `_instruments` and `_covar` attributes, respectively.
        It splits these strings by '+' to identify individual variables. If any
        variable is found to be listed as both an instrument and a covariate,
        an `InstrumentsAsCovarsError` is raised, indicating that
        instrument variables cannot be specified as covariates.

        Raises
        ------
        InstrumentsAsCovarsError
            If any instrument variables are also specified as covariates.
            The error message includes the names of the variables causing
            this conflict.

        Notes
        -----
        - The `_instruments` and `_covar` attributes must be strings containing variable
          names separated by '+'. If either attribute is `None`, this check is skipped.
        - This method does not return any value but raises an error if the check fails.
        """
        instruments = self._instruments
        covars = self._covar

        if instruments is not None:
            instruments_as_covars = [
                element
                for element in instruments.split("+")
                if element in covars.split("+")
            ]

            if instruments_as_covars:
                raise InstrumentsAsCovarsError(
                    f"""
                    The instrument(s) {",".join(instruments_as_covars)} are specified as
                    covariates in the first part of the three-part formula. This is not allowed.
                    """
                )


def _get_first_and_second_stage_fml(
    depvar: str,
    covar: str,
    fval: str,
    endogvar: Optional[str] = None,
    instruments: Optional[str] = None,
) -> tuple:
    """
    Generate first and second stage formulas for OLS and IV regression models.

    Parameters
    ----------
    depvar : str
        Dependent variable.
    covar : str
        Covariates.
    fval : str, optional
        Fixed effect variable, defaults to "0" if None.
    endogvar : str, optional
        Endogenous variable.
    instruments : str
        Instrumental variables.

    Returns
    -------
    tuple
        (Second stage formula, First stage formula).

    Notes
    -----
    First stage formula is None if `endogvar` is not provided.
    """
    if fval is None:
        fval = "0"

    fml_iv = f"{endogvar} ~ {instruments}"

    fml_second_stage = f"{depvar} ~ {covar} + 1"
    fml_first_stage = f"{fml_iv}+{covar}-{endogvar} + 1" if endogvar else None

    return fml_second_stage, fml_first_stage


def collect_fml_dict(
    fevars_formula: list[str], depvars_dict: list[str], covars_formula: list[str]
):
    """
    Create a nested dictionary mapping fixed effects to depvar-covariate formulas.

    Parameters
    ----------
    fevars_formula : list
        Fixed effects variables formulas.
    depvars_dict : dict
        Mapping of dependent variables to their properties.
    covars_formula : list
        Covariate formulas.

    Returns
    -------
    dict
        Nested dict of fevar to depvar to list of covariate formulas.
    """
    fml_dict: dict[str, dict[str, list[str]]] = {}

    for fevar in fevars_formula:
        res: dict[str, list[str]] = {}
        for depvar in depvars_dict:
            res[depvar] = []
            for covar in covars_formula:
                res[depvar].append(f"{depvar}~{covar}")
        fml_dict[fevar] = res

    return fml_dict


def _deparse_fml(
    fml: str,
) -> tuple[str, str, str, Union[str, None], Union[str, None]]:
    """
    Decompose a formula string into its constituent parts.

    This function takes a formula string and splits it into its components based on
    the presence of '~' and '|' characters. The formula string is expected to follow
    the format 'depvars ~ covars | fevars | endogvars ~ instruments'.

    Parameters
    ----------
    fml : str
        The formula string to be decomposed.

    Returns
    -------
    tuple
        A tuple containing the decomposed parts of the formula as strings:
        (depvars, covars, fevars,endogvars, instruments).
        `endogvars` and `instruments` may be `None` if not applicable.

    Raises
    ------
    UnderDeterminedIVError
        If the number of instruments is less than the number of endogenous variables,
        indicating an underdetermined IV system.

    Notes
    -----
    - The function cleans the formula string of spaces before processing.
    - Fixed effects variables are set to "0" if not explicitly provided in
      the formula.
    - The function automatically adds endogenous variables to the covariates
      list, a behavior indicated by the comment on potentially misleading
      naming conventions.

    Examples
    --------
    ```{python}
    fml = "y ~ x1+x2|z1+z2|w1 ~ w2+w3"
    _deparse_fml(fml)
    ('y', 'w1+x1+x2', 'z1+z2', 'w1', 'w2+w3')
    ```

    Here, `y` is the dependent variable, `x1+x2` are the covariates
    (with `w1` added due to it being an endogenous variable), `z1+z2` are
    the fixed effects variables, and `w1` is the endogenous
    variable with `w2+w3` as its instruments.
    """
    # Clean up the formula string
    fml = "".join(fml.split())

    # Split the formula string into its components
    fml_split = fml.split("|")
    depvars, covars = fml_split[0].split("~")

    if len(fml_split) == 1:
        fevars = "0"
        endogvars = None
        instruments = None
    elif len(fml_split) == 2:
        if "~" in fml_split[1]:
            fevars = "0"
            endogvars, instruments = fml_split[1].split("~")
            # add endogenous variable to "covars" - yes, bad naming
            _check_endogvars_as_covars(endogvars, covars)
            covars = endogvars if covars == "1" else f"{endogvars}+{covars}"
        else:
            fevars = fml_split[1]
            endogvars = None
            instruments = None
    elif len(fml_split) == 3:
        fevars = fml_split[1]
        endogvars, instruments = fml_split[2].split("~")
        _check_endogvars_as_covars(endogvars, covars)

        # add endogenous variable to "covars" - yes, bad naming
        covars = endogvars if covars == "1" else f"{endogvars}+{covars}"

    endogvars_list = []
    instruments_list = []

    if endogvars is not None and not isinstance(endogvars, list):
        endogvars_list = endogvars.split("+")

    if instruments is not None and not isinstance(instruments, list):
        instruments_list = instruments.split("+")

    if endogvars_list and instruments_list:
        if len(endogvars_list) > len(instruments_list):
            raise UnderDeterminedIVError(
                """
                The IV system is underdetermined. Please provide as many or
                more instruments as endogenous variables.
                """
            )
        else:
            pass

    return depvars, covars, fevars, endogvars, instruments


def _check_endogvars_as_covars(endogvars: str, covars: str):
    """
    Check if one or more endogenous variables are included in the covariates.

    Parameters
    ----------
    endogvars : str
        A string representing the endogenous variables in the model,
        separated by "+".
    covars : str
        A string representing the covariates in the model, separated by "+".

    Raises
    ------
    EndogVarsAsCovarsError
        If any of the specified endogenous variables are also listed as covariates.

    Returns
    -------
    None
    """
    endogvars_as_covars = [
        element for element in endogvars.split("+") if element in covars.split("+")
    ]

    if endogvars_as_covars:
        raise EndogVarsAsCovarsError(
            f"""
                The endogeneous variable(s) {",".join(endogvars_as_covars)} are specified as
                covariates in the first part of the three-part formula. This is not allowed.
                """
        )


def _input_formula_to_dict(x: str) -> dict[str, list[str]]:
    """
    Parse a formula string.

    Given a formula string `x` - e.g. 'X1 + csw(X2, X3)' - , splits it into its
    constituent variables and their types (if any), and returns a dictionary
    containing the result. The dictionary has the following keys: 'constant',
    'sw', 'sw0', 'csw'.

    The values are lists of variables of the respective type.

    Parameters
    ----------
    x : str
        The formula string to unpack.

    Returns
    -------
    res_s : dict
        A dictionary containing the unpacked formula. The dictionary has the
        following keys:
            - 'constant' : list of str
                The list of constant (i.e., non-switched) variables in the formula.
            - 'sw' : list of str
                The list of variables that have a regular switch
                (i.e., 'sw(var1, var2, ...)' notation) in the formula.
            - 'sw0' : list of str
                The list of variables that have a 'sw0(var1, var2, ..)' switch
                in the formula.
            - 'csw' : list of str or list of lists of str
                The list of variables that have a 'csw(var1, var2, ..)' switch
                in the formula.
                Each element in the list can be either a single variable string,
                or a list of variable strings if multiple variables are listed
                in the switch.
            - 'csw0' : list of str or list of lists of str
                The list of variables that have a 'csw0(var1,var2,...)' switch
                in the formula.
                Each element in the list can be either a single variable string,
                or a list of variable strings if multiple variables are listed
                in the switch.

    Raises
    ------
    ValueError:
        If the switch type is not one of 'sw', 'sw0', 'csw', or 'csw0'.

    Example:
    --------
    >>> _input_formula_to_dict("a+sw(b)+csw(x1,x2)+sw0(d)+csw0(y1,y2,y3)")
    {'constant': ['a'],
     'sw': ['b'],
     'csw': [['x1', 'x2']],
     'sw0': ['d'],
     'csw0': [['y1', 'y2', 'y3']]}
    """
    # Split the formula into its constituent variables
    var_split = x.split("+")

    res_s: dict[str, list[str]] = {"constant": []}
    for var in var_split:
        # Check if this variable contains a switch
        varlist, sw_type = _find_multiple_estimation_syntax(var)

        # If there's no switch, just add the variable to the list
        if sw_type is None:
            res_s["constant"].append(varlist)

        # If there'_ a switch, unpack it and add it to the list
        else:
            if sw_type in ["sw", "sw0", "csw", "csw0"]:
                _check_duplicate_key(res_s, sw_type)
                res_s[sw_type] = varlist
            elif sw_type == "varying_slopes":
                res_s[sw_type] = varlist
            else:
                raise UnsupportedMultipleEstimationSyntax("Unsupported switch type")

    # Sort the list by type (strings first, then lists)
    # res_s.sort(key=lambda x: 0 if isinstance(x, str) else 1)

    return res_s


def _dict_to_list_of_formulas(unpacked: dict[str, list[str]]) -> list[str]:
    """
    Generate a list of formula strings from a dictionary of "unpacked" fml vars.

    Given a dictionary of "unpacked" formula variables, returns a string containing
    formulas. An "unpacked" formula is a deparsed formula that allows for multiple
    estimations.

    Parameters
    ----------
    unpacked : dict
        A dictionary of unpacked formula variables. The dictionary has the following
        keys:
            - 'constant' : list of str
                The list of constant (i.e., non-switched) variables in the formula.
            - 'sw' : list of str
                The list of variables that have a regular switch
                (i.e., 'sw(var1, var2, ...)' notation) in the formula.
            - 'sw0' : list of str
                The list of variables that have a 'sw0(var1, var2, ..)' switch
                in the formula.
            - 'csw' : list of str or list of lists of str
                The list of variables that have a 'csw(var1, var2, ..)' switch
                in the formula.
                Each element in the list can be either a single variable string,
                or a list of variable strings if multiple variables are listed
                in the switch.
            - 'csw0' : list of str or list of lists of str
    """
    res = {"constant": unpacked.get("constant", [])}

    # add up all variable constants (only required for csw)
    if "csw" in unpacked:
        res["variable"] = unpacked["csw"]
        variable_type = "csw"
    elif "csw0" in unpacked:
        res["variable"] = unpacked["csw0"]
        variable_type = "csw0"
    elif "sw" in unpacked:
        res["variable"] = unpacked["sw"]
        variable_type = "sw"
    elif "sw0" in unpacked:
        res["variable"] = unpacked["sw0"]
        variable_type = "sw0"
    else:
        res["variable"] = []
        variable_type = None

    const_fml = "+".join(res["constant"]) if res["constant"] else ""

    variable_fml = []
    if res["variable"]:
        if variable_type in ["csw", "csw0"]:
            variable_fml = [
                "+".join(res["variable"][: i + 1]) for i in range(len(res["variable"]))
            ]
        else:
            variable_fml = [res["variable"][i] for i in range(len(res["variable"]))]
        if variable_type in ["sw0", "csw0"]:
            variable_fml = ["0", *variable_fml]

    fml_list = []
    if variable_fml:
        if const_fml:
            fml_list = [
                f"{const_fml}+{variable_fml[i]}"
                for i in range(len(variable_fml))
                if variable_fml[i] != "0"
            ]
            if variable_type in ["sw0", "csw0"]:
                fml_list.insert(0, const_fml)
        else:
            fml_list = variable_fml
    elif const_fml:
        fml_list.append(const_fml)
    else:
        raise AttributeError("Not a valid formula provided.")

    if not isinstance(fml_list, list):
        fml_list = [fml_list]

    return fml_list


def _find_multiple_estimation_syntax(x: str):
    """
    Search for matches of multiple estimation syntax in a string.

    Matches are either 'sw', 'sw0', 'csw', 'csw0'. If a match is found, returns a
    tuple containing a list of the elements found and the type of match. Otherwise,
    returns the original string and None.

    Parameters
    ----------
    x : str
        The string to search for matches in.

    Returns
    -------
    list[str] or str, str or None
        If any matches were found, returns a
        tuple containing a list of the elements found and the type of match
        (either 'sw', 'sw0', 'csw', or 'csw0').
        Otherwise, returns the original string and None.

    Example:
        _find_multiple_estimation_syntax('sw(var1, var2)') -> (['var1', ' var2'], 'sw')
    """
    # Search for matches in the string
    sw_match = re.findall(r"sw\((.*?)\)", x)
    csw_match = re.findall(r"csw\((.*?)\)", x)
    sw0_match = re.findall(r"sw0\((.*?)\)", x)
    csw0_match = re.findall(r"csw0\((.*?)\)", x)

    # Check for sw matches
    if sw_match:
        if csw_match:
            return csw_match[0].split(","), "csw"
        else:
            return sw_match[0].split(","), "sw"

    # Check for sw0 matches
    elif sw0_match:
        if csw0_match:
            return csw0_match[0].split(","), "csw0"
        else:
            return sw0_match[0].split(","), "sw0"

    # No matches found
    else:
        return x, None


def _check_duplicate_key(my_dict: dict, key: str):
    """
    Identify duplicate keys.

    Checks if a key already exists in a dictionary. If it does, raises a
    DuplicateKeyError. Otherwise, does nothing.

    Parameters
    ----------
    my_dict : dict
        The dictionary to check for duplicate keys.
    key : str
        The key to check for in the dictionary.

    Returns
    -------
        None
    """
    for key in ["sw", "csw", "sw0", "csw0"]:
        if key in my_dict:
            raise DuplicateKeyError(
                f"""
                Duplicate key found: "{key}. Multiple estimation syntax can
                only be used once on the rhs of the two-sided formula.
                """
            )



import pandas as pd
import pytest
import statsmodels.formula.api as smf

import pyfixest as pf
from pyfixest.estimation.estimation import feols, fepois
from pyfixest.report.summarize import _select_order_coefs, etable, summary
from pyfixest.utils.utils import get_data


def test_summary():
    """Just run etable() and summary() on a few models."""
    df1 = get_data()
    df1 = pd.concat(
        [df1, df1], axis=0
    )  # Make it a bit larger, for examining the thousands separator
    df2 = get_data(model="Fepois")

    fit1 = feols("Y ~ X1 + X2 | f1", data=df1)
    fit1a = feols("Y ~ X1 + X2 + f1", data=df1)
    fit2 = fepois("Y ~ X1 + X2 + f2 | f1", data=df2, vcov={"CRV1": "f1+f2"})
    fit3 = feols("Y ~ X1", data=df1)
    fit4 = feols("Y ~ X1", data=df1, weights="weights")
    fit5 = feols("Y ~ 1 | Z1 ~ X1", data=df1)

    fit_qreg = pf.quantreg("Y ~ X1", data=df1, vcov="nid")

    summary(fit1)
    summary(fit2)
    summary([fit1, fit2])
    summary([fit4])
    fit5.summary()

    etable(fit1)
    etable(fit2)
    etable([fit1, fit2])

    etable([fit3])
    etable([fit1, fit2, fit3])

    fit_iv = feols("Y ~ X2 | f1 | X1 ~ Z1", data=df1)
    etable([fit_iv, fit1])

    fit_multi = feols("Y + Y2 ~ X1 + X2 | f1", data=df1)
    etable(fit_multi.to_list())

    # Test significance code
    etable([fit1, fit2], signif_code=[0.01, 0.05, 0.1])
    etable([fit1, fit2], signif_code=[0.02, 0.06, 0.1])

    # Test coefficient format
    etable([fit1, fit2], coef_fmt="b (se)\nt [p]")

    # Test custom statistics
    etable(
        models=[fit1, fit2],
        custom_stats={
            "conf_int_lb": [fit1._conf_int[0], fit2._conf_int[0]],
            "conf_int_ub": [fit1._conf_int[1], fit2._conf_int[1]],
        },
        coef_fmt="b [conf_int_lb, conf_int_ub]",
    )

    # Test scientific notation
    etable(
        models=[fit1],
        custom_stats={
            "test_digits": [[0.1, 12300]],
        },
        coef_fmt="b [test_digits]",
        digits=2,
    )

    # Test scientific notation, thousands separator
    etable(
        models=[fit1],
        custom_stats={
            "test_digits": [[0.1, 12300]],
        },
        coef_fmt="b [test_digits]",
        digits=2,
        scientific_notation=False,
        thousands_sep=True,
    )

    # Test select / order coefficients
    etable([fit1, fit2, fit3], coef_fmt="b (se)\nt [p]")
    etable([fit1, fit2, fit3], coef_fmt="b (se)\nt [p]", keep=["X1", "cep"])
    etable([fit1, fit2, fit3], coef_fmt="b (se)\nt [p]", drop=[r"\d$"])
    etable([fit1, fit2, fit3], coef_fmt="b (se)\nt [p]", keep=[r"\d"], drop=["f"])
    etable([fit1, fit2, fit3], coef_fmt="b (se)\nt [p]", keep="X")
    etable([fit1, fit2, fit3], coef_fmt="b (se)\nt [p]", drop=r"\d$")

    # test labels, felabels args
    etable([fit1, fit1a], labels={"X1": "X1_label"}, felabels={"f1": "f1_label"})
    etable(
        [fit1, fit1a], labels={"X1": "X1_label"}, felabels={"f1": "f1_label"}, keep="X1"
    )
    etable(
        [fit1, fit1a], labels={"X1": "X1_label"}, felabels={"f1": "f1_label"}, drop="X1"
    )
    etable([fit1, fit1a], felabels={"f1": "f1_renamed2"}, keep=["f1"])

    cols = ["x1", "x2", "x11", "x21"]
    assert _select_order_coefs(cols, keep=["x1"]) == ["x1", "x11"]
    assert _select_order_coefs(cols, drop=["x1"]) == ["x2", "x21"]
    assert _select_order_coefs(cols, keep=["x1"], exact_match=True) == ["x1"]
    assert _select_order_coefs(cols, drop=["x1"], exact_match=True) == [
        "x2",
        "x11",
        "x21",
    ]

    # API tests for new tex args

    etable([fit1, fit2], type="tex")
    etable([fit1, fit2], type="tex", print_tex=True)

    etable([fit1, fit2], type="tex", notes="You can add notes here.")
    etable([fit1, fit2], type="md", notes="You can add notes here.")

    etable([fit1, fit2], type="tex", model_heads=["Model 1", "Model 2"])
    etable(
        [fit1, fit2], type="tex", model_heads=["Model 1", "Model 2"], head_order="dh"
    )
    etable(
        [fit1, fit2], type="tex", model_heads=["Model 1", "Model 2"], head_order="hd"
    )
    etable([fit1, fit2], type="tex", model_heads=["Model 1", "Model 2"], head_order="d")
    etable([fit1, fit2], type="tex", model_heads=["Model 1", "Model 2"], head_order="h")
    etable([fit1, fit2], type="tex", model_heads=["Model 1", "Model 2"], head_order="")
    etable([fit1, fit2], type="tex", filename="tests/texfiles/test.tex")

    summary(fit_qreg)
    etable(fit_qreg)


@pytest.mark.skip("Pyfixest PR is not yet merged into stargazer.")
def test_stargazer():
    data = pf.get_data()

    fit = pf.feols("Y ~ X1", data=data)
    fit_smf = smf.ols("Y ~ X1", data=data).fit()

    pf.Stargazer([fit, fit_smf])
